@@ -124,6 +124,7 @@ type token =
 	| IntInterval of string
 	| Macro of string
 	| Question
+	| At
 
 type unop_flag =
 	| Prefix
@@ -133,7 +134,7 @@ type while_flag =
 	| NormalWhile
 	| DoWhile
 
-type type_path_normal = {
+type type_path = {
 	tpackage : string list;
 	tname : string;
 	tparams : type_param_or_const list;
@@ -141,24 +142,24 @@ type type_path_normal = {
 }
 
 and type_param_or_const =
-	| TPType of type_path
+	| TPType of complex_type
 	| TPConst of constant
 
 and anonymous_field =
-	| AFVar of type_path
-	| AFProp of type_path * string * string
-	| AFFun of (string * bool * type_path) list * type_path
+	| AFVar of complex_type
+	| AFProp of complex_type * string * string
+	| AFFun of (string * bool * complex_type) list * complex_type
 
-and type_path =
-	| TPNormal of type_path_normal
-	| TPFunction of type_path list * type_path
-	| TPAnonymous of (string * bool option * anonymous_field * pos) list
-	| TPParent of type_path
-	| TPExtend of type_path_normal * (string * bool option * anonymous_field * pos) list
+and complex_type =
+	| CTPath of type_path
+	| CTFunction of complex_type list * complex_type
+	| CTAnonymous of (string * bool option * anonymous_field * pos) list
+	| CTParent of complex_type
+	| CTExtend of type_path * (string * bool option * anonymous_field * pos) list
 
 type func = {
-	f_args : (string * bool * type_path option * expr option) list;
-	f_type : type_path option;
+	f_args : (string * bool * complex_type option * expr option) list;
+	f_type : complex_type option;
 	f_expr : expr;
 }
 
@@ -172,32 +173,34 @@ and expr_def =
 	| EObjectDecl of (string * expr) list
 	| EArrayDecl of expr list
 	| ECall of expr * expr list
-	| ENew of type_path_normal * expr list
+	| ENew of type_path * expr list
 	| EUnop of unop * unop_flag * expr
-	| EVars of (string * type_path option * expr option) list
-	| EFunction of func
+	| EVars of (string * complex_type option * expr option) list
+	| EFunction of string option * func
 	| EBlock of expr list
 	| EFor of string * expr * expr
 	| EIf of expr * expr * expr option
 	| EWhile of expr * expr * while_flag
 	| ESwitch of expr * (expr list * expr) list * expr option
-	| ETry of expr * (string * type_path * expr) list
+	| ETry of expr * (string * complex_type * expr) list
 	| EReturn of expr option
 	| EBreak
 	| EContinue
 	| EUntyped of expr
 	| EThrow of expr
-	| ECast of expr * type_path option
-	| EDisplay of expr
-	| EDisplayNew of type_path_normal
+	| ECast of expr * complex_type option
+	| EDisplay of expr * bool
+	| EDisplayNew of type_path
 	| ETernary of expr * expr * expr
 
 and expr = expr_def * pos
 
-type type_param = string * type_path_normal list
+type type_param = string * type_path list
 
 type documentation = string option
 
+type metadata = (string * expr list * pos) list
+
 type access =
 	| APublic
 	| APrivate
@@ -206,10 +209,19 @@ type access =
 	| ADynamic
 	| AInline
 
-type class_field =
-	| FVar of string * documentation * access list * type_path option * expr option
-	| FFun of string * documentation * access list * type_param list * func
-	| FProp of string * documentation * access list * string * string * type_path
+type class_field_kind =
+	| FVar of complex_type option * expr option
+	| FFun of type_param list * func
+	| FProp of string * string * complex_type
+
+type class_field = {
+	cff_name : string;
+	cff_doc : documentation;
+	cff_pos : pos;
+	mutable cff_meta : metadata;
+	mutable cff_access : access list;
+	mutable cff_kind : class_field_kind;
+}
 
 type enum_flag =
 	| EPrivate
@@ -219,25 +231,26 @@ type class_flag =
 	| HInterface
 	| HExtern
 	| HPrivate
-	| HExtends of type_path_normal
-	| HImplements of type_path_normal
+	| HExtends of type_path
+	| HImplements of type_path
 
-type enum_constructor = string * documentation * (string * bool * type_path) list * pos
+type enum_constructor = string * documentation * metadata * (string * bool * complex_type) list * pos
 
 type ('a,'b) definition = {
 	d_name : string;
 	d_doc : documentation;
 	d_params : type_param list;
+	d_meta : metadata;
 	d_flags : 'a list;
 	d_data : 'b;
 }
 
 type type_def =
-	| EClass of (class_flag, (class_field * pos) list) definition
+	| EClass of (class_flag, class_field list) definition
 	| EEnum of (enum_flag, enum_constructor list) definition
-	| ETypedef of (enum_flag, type_path) definition
-	| EImport of type_path_normal
-	| EUsing of type_path_normal
+	| ETypedef of (enum_flag, complex_type) definition
+	| EImport of type_path
+	| EUsing of type_path
 
 type type_decl = type_def * pos
 
@@ -255,7 +268,7 @@ let is_prefix = function
 
 let base_class_name = snd
 
-let null_pos = { pfile = "<null>"; pmin = -1; pmax = -1 }
+let null_pos = { pfile = "?"; pmin = -1; pmax = -1 }
 
 let punion p p2 =
 	{
@@ -266,6 +279,11 @@ let punion p p2 =
 
 let s_type_path (p,s) = match p with [] -> s | _ -> String.concat "." p ^ "." ^ s
 
+let parse_path s =
+	match List.rev (ExtString.String.nsplit s ".") with
+	| [] -> failwith "Invalid empty path"
+	| x :: l -> List.rev l, x
+
 let s_escape s =
 	let b = Buffer.create (String.length s) in
 	for i = 0 to (String.length s) - 1 do
@@ -287,6 +305,14 @@ let s_constant = function
 	| Type s -> s
 	| Regexp (r,o) -> "~/" ^ r ^ "/"
 
+let s_access = function
+	| APublic -> "public"
+	| APrivate -> "private"
+	| AStatic -> "static"
+	| AOverride -> "override"
+	| ADynamic -> "dynamic"
+	| AInline -> "inline"
+
 let s_keyword = function
 	| Function -> "function"
 	| Class -> "class"
@@ -380,6 +406,7 @@ let s_token = function
 	| IntInterval s -> s ^ "..."
 	| Macro s -> "#" ^ s
 	| Question -> "?"
+	| At -> "@"
 
 let unescape s =
 	let b = Buffer.create 0 in
