@@ -33,6 +33,25 @@ let type_inline ctx cf f ethis params tret p =
 		Hashtbl.add hcount name (ref 0);
 		(name,t)
 	) f.tf_args in
+	(* type substitution on both class and function type parameters *)
+	let has_params, map_type =
+		let rec get_params c pl =
+			match c.cl_super with
+			| None -> c.cl_types, pl
+			| Some (csup,spl) ->
+				let spl = (match apply_params c.cl_types pl (TInst (csup,spl)) with
+				| TInst (_,pl) -> pl
+				| _ -> assert false
+				) in
+				let ct, cpl = get_params csup spl in
+				c.cl_types @ ct, pl @ cpl
+		in
+		let tparams = (match follow ethis.etype with TInst (c,pl) -> get_params c pl | _ -> ([],[])) in
+		let pmonos = List.map (fun _ -> mk_mono()) cf.cf_params in
+		let tmonos = snd tparams @ pmonos in
+		let tparams = fst tparams @ cf.cf_params in
+		tparams <> [], apply_params tparams tmonos
+	in
 	(* use default values for null/unset arguments *)
 	let rec loop pl al =
 		match pl, al with
@@ -40,12 +59,12 @@ let type_inline ctx cf f ethis params tret p =
 		| e :: pl, (name, opt, t) :: al ->
 			if is_nullable t && is_null e.etype then Hashtbl.add lsets name (); (* force coerce *)
 			(match e.eexpr, opt with
-			| TConst TNull , Some c -> mk (TConst c) t e.epos
+			| TConst TNull , Some c -> mk (TConst c) (map_type t) e.epos
 			| _ -> e) :: loop pl al
 		| [], (_,opt,t) :: al ->
 			(match opt with
 			| None -> assert false
-			| Some c -> mk (TConst c) t p) :: loop [] al
+			| Some c -> mk (TConst c) (map_type t) p) :: loop [] al
 		| _ :: _, [] ->
 			assert false
 	in
@@ -89,15 +108,31 @@ let type_inline ctx cf f ethis params tret p =
 			| None -> mk (TConst TNull) (mk_mono()) p
 			| Some e -> map term e)
 		| TFor (v,t,e1,e2) ->
-			{ e with eexpr = TFor (local v,t,map false e1,map false e2) }
+			let e1 = map false e1 in
+			let old = save_locals ctx in
+			let v = add_local ctx v t in
+			let e2 = map false e2 in
+			old();
+			{ e with eexpr = TFor (v,t,e1,e2) }
 		| TMatch (e,en,cases,def) ->
-			let term, t = (match def with Some d when term -> true, d.etype | _ -> false, e.etype) in
+			let term, t = (match def with Some d when term -> true, ref d.etype | _ -> false, ref e.etype) in
 			let cases = List.map (fun (i,vl,e) ->
-				i, opt (List.map (fun (n,t) -> opt local n, t)) vl, map term e
+				let old = save_locals ctx in
+				let vl = opt (List.map (fun (n,t) -> opt (fun n -> add_local ctx n t) n, t)) vl in
+				let e = map term e in
+				if is_null e.etype then t := e.etype;
+				old();
+				i, vl, e
 			) cases in
-			{ e with eexpr = TMatch (map false e,en,cases,opt (map term) def); etype = t }
+			{ e with eexpr = TMatch (map false e,en,cases,opt (map term) def); etype = !t }
 		| TTry (e1,catches) ->
-			{ e with eexpr = TTry (map term e1,List.map (fun (v,t,e) -> local v,t,map term e) catches) }
+			{ e with eexpr = TTry (map term e1,List.map (fun (v,t,e) ->
+				let old = save_locals ctx in
+				let v = add_local ctx v t in
+				let e = map term e in
+				old();
+				v,t,e
+			) catches) }
 		| TBlock l ->
 			let old = save_locals ctx in
 			let t = ref e.etype in
@@ -121,17 +156,17 @@ let type_inline ctx cf f ethis params tret p =
 			let econd = map false econd in
 			let eif = map term eif in
 			let eelse = map term eelse in
-			{ e with eexpr = TIf(econd,eif,Some eelse); etype = eif.etype }
+			{ e with eexpr = TIf(econd,eif,Some eelse); etype = if is_null eif.etype then eif.etype else eelse.etype }
 		| TParenthesis _ | TIf (_,_,Some _) | TSwitch (_,_,Some _) ->
 			Type.map_expr (map term) e
 		| TUnop (op,pref,({ eexpr = TLocal s } as e1)) ->
 			(match op with
-			| Increment | Decrement -> Hashtbl.add lsets s ()
+			| Increment | Decrement -> Hashtbl.add lsets (local s) ()
 			| _ -> ());
 			{ e with eexpr = TUnop (op,pref,map false e1) }
 		| TBinop (op,({ eexpr = TLocal s } as e1),e2) ->
 			(match op with
-			| OpAssign | OpAssignOp _ -> Hashtbl.add lsets s ()
+			| OpAssign | OpAssignOp _ -> Hashtbl.add lsets (local s) ()
 			| _ -> ());
 			{ e with eexpr = TBinop (op,map false e1,map false e2) }
 		| TConst TSuper ->
@@ -147,7 +182,7 @@ let type_inline ctx cf f ethis params tret p =
 	Hashtbl.add hcount vthis this_count;
 	let vars = List.map2 (fun (n,t) e ->
 		let flag = not (Hashtbl.mem lsets n) && (match e.eexpr with
-			| TLocal _ | TConst _ -> true
+			| TLocal _ | TConst _ | TFunction _ -> true
 			| _ ->
 				let used = !(Hashtbl.find hcount n) in
 				used <= 1
@@ -168,31 +203,33 @@ let type_inline ctx cf f ethis params tret p =
 		| _ -> Type.map_expr inline_params e
 	in
 	let e = (if PMap.is_empty subst then e else inline_params e) in
-	let init = (match vars with [] -> None | l -> Some (mk (TVars (List.rev l)) ctx.api.tvoid p)) in
+	let init = (match vars with [] -> None | l -> Some (mk (TVars (List.rev l)) ctx.t.tvoid p)) in
 	if Common.defined ctx.com "js" && (init <> None || !has_vars) then
 		None
 	else
+		let wrap e =
+			(* we can't mute the type of the expression because it is not correct to do so *)
+			if e.etype == tret then
+				e
+			else
+				mk (TParenthesis e) tret e.epos
+		in
 		let e = (match e.eexpr, init with
-			| TBlock [e] , None -> { e with etype = tret; }
-			| _ , None -> { e with etype = tret; }
+			| TBlock [e] , None -> wrap e
+			| _ , None -> wrap e
 			| TBlock l, Some init -> mk (TBlock (init :: l)) tret e.epos
 			| _, Some init -> mk (TBlock [init;e]) tret e.epos
 		) in
 		(* we need to replace type-parameters that were used in the expression *)
-		let tparams = (match follow ethis.etype with TInst (c,pl) -> (c.cl_types,pl) | _ -> ([],[])) in
-		match cf.cf_params, tparams with
-		| [], ([],_) -> Some e
-		| _ ->
-			let pmonos = List.map (fun _ -> mk_mono()) cf.cf_params in
-			let tmonos = snd tparams @ pmonos in
-			let tparams = fst tparams @ cf.cf_params in
-			let mt = apply_params tparams tmonos cf.cf_type in
+		if not has_params then
+			Some e
+		else
+			let mt = map_type cf.cf_type in
 			unify_raise ctx mt (TFun (List.map (fun e -> "",false,e.etype) params,tret)) p;
 			(*
 				this is very expensive since we are building the substitution list for
 				every expression, but hopefully in such cases the expression size is small
 			*)
-			let map_type t = apply_params tparams tmonos t in
 			let rec map_expr_type e = Type.map_expr_type map_expr_type map_type e in
 			Some (map_expr_type e)
 
@@ -200,8 +237,8 @@ let type_inline ctx cf f ethis params tret p =
 (* LOOPS *)
 
 let optimize_for_loop ctx i e1 e2 p =
-	let t_void = ctx.api.tvoid in
-	let t_int = ctx.api.tint in
+	let t_void = ctx.t.tvoid in
+	let t_int = ctx.t.tint in
 	let lblock el = Some (mk (TBlock el) t_void p) in
 	match e1.eexpr, follow e1.etype with
 	| TNew ({ cl_path = ([],"IntIter") },[],[i1;i2]) , _ ->
@@ -234,21 +271,24 @@ let optimize_for_loop ctx i e1 e2 p =
 			| TBlock el -> mk (TBlock (init :: el)) t_void e2.epos
 			| _ -> mk (TBlock [init;e2]) t_void p
 		in
+		(*
+			force locals to be of Int type (to prevent Int/UInt issues)
+		*)
 		(match max with
 		| None ->
 			lblock [
-				mk (TVars [tmp,i1.etype,Some i1]) t_void p;
+				mk (TVars [tmp,t_int,Some i1]) t_void p;
 				mk (TWhile (
-					mk (TBinop (OpLt, etmp, i2)) ctx.api.tbool p,
+					mk (TBinop (OpLt, etmp, { i2 with etype = t_int })) ctx.t.tbool p,
 					block,
 					NormalWhile
 				)) t_void p;
 			]
 		| Some max ->
 			lblock [
-				mk (TVars [tmp,i1.etype,Some i1;max,i2.etype,Some i2]) t_void p;
+				mk (TVars [tmp,t_int,Some i1;max,t_int,Some i2]) t_void p;
 				mk (TWhile (
-					mk (TBinop (OpLt, etmp, mk (TLocal max) i2.etype p)) ctx.api.tbool p,
+					mk (TBinop (OpLt, etmp, mk (TLocal max) t_int p)) ctx.t.tbool p,
 					block,
 					NormalWhile
 				)) t_void p;
@@ -275,7 +315,7 @@ let optimize_for_loop ctx i e1 e2 p =
 		lblock [
 			mk (TVars (ivar :: avars)) t_void p;
 			mk (TWhile (
-				mk (TBinop (OpLt, iexpr, mk (TField (arr,"length")) t_int p)) ctx.api.tbool p,
+				mk (TBinop (OpLt, iexpr, mk (TField (arr,"length")) t_int p)) ctx.t.tbool p,
 				block,
 				NormalWhile
 			)) t_void p;
@@ -295,7 +335,7 @@ let optimize_for_loop ctx i e1 e2 p =
 		lblock [
 			mk (TVars [cell,tcell,Some (mk (TField (e1,"head")) tcell p)]) t_void p;
 			mk (TWhile (
-				mk (TBinop (OpNotEq, cexpr, mk (TConst TNull) tcell p)) ctx.api.tbool p,
+				mk (TBinop (OpNotEq, cexpr, mk (TConst TNull) tcell p)) ctx.t.tbool p,
 				block,
 				NormalWhile
 			)) t_void p
@@ -303,17 +343,129 @@ let optimize_for_loop ctx i e1 e2 p =
 	| _ ->
 		None
 
+(* ---------------------------------------------------------------------- *)
+(* SANITIZE *)
+
+(*
+	makes sure that when an AST get generated to source code, it will not
+	generate expressions that evaluate differently. It is then necessary to
+	add parenthesises around some binary expressions when the AST does not
+	correspond to the natural operand priority order for the platform
+*)
+
+(*
+	this is the standard C++ operator precedence, which is also used by both JS and PHP
+*)
+let standard_precedence op =
+	let left = true and right = false in
+	match op with
+	| OpMult | OpDiv | OpMod -> 5, left
+	| OpAdd | OpSub -> 6, left
+	| OpShl | OpShr | OpUShr -> 7, left
+	| OpLt | OpLte | OpGt | OpGte -> 8, left
+	| OpEq | OpNotEq -> 9, left
+	| OpAnd -> 10, left
+	| OpXor -> 11, left
+	| OpOr -> 12, left
+	| OpInterval -> 13, right (* haxe specific *)
+	| OpBoolAnd -> 14, left
+	| OpBoolOr -> 15, left
+	| OpAssignOp OpAssign -> 16, right (* mimics ?: *)
+	| OpAssign | OpAssignOp _ -> 17, right
+
+let sanitize_expr e =
+	let parent e =
+		mk (TParenthesis e) e.etype e.epos
+	in
+	let block e =
+		mk (TBlock [e]) e.etype e.epos
+	in
+	let need_parent e =
+		match e.eexpr with
+		| TConst _ | TLocal _ | TEnumField _ | TArray _ | TField _ | TParenthesis _ | TCall _ | TClosure _ | TNew _ | TTypeExpr _ | TObjectDecl _ | TArrayDecl _ -> false
+		| TCast _ | TThrow _ | TReturn _ | TTry _ | TMatch _ | TSwitch _ | TFor _ | TIf _ | TWhile _ | TBinop _ | TContinue | TBreak
+		| TBlock _ | TVars _ | TFunction _ | TUnop _ -> true
+	in
+	match e.eexpr with
+	| TBinop (op,e1,e2) ->
+		let swap op1 op2 =
+			let p1, left1 = standard_precedence op1 in
+			let p2, _ = standard_precedence op2 in
+			left1 && p1 <= p2
+		in
+		let rec loop ee left =
+			match ee.eexpr with
+			| TBinop (op2,_,_) -> if left then not (swap op2 op) else swap op op2
+			| TIf _ -> if left then not (swap (OpAssignOp OpAssign) op) else swap op (OpAssignOp OpAssign)
+			| TCast (e,None) -> loop e left
+			| _ -> false
+		in		
+		let e1 = if loop e1 true then parent e1 else e1 in
+		let e2 = if loop e2 false then parent e2 else e2 in
+		{ e with eexpr = TBinop (op,e1,e2) }
+	| TUnop (op,mode,e2) ->
+		let rec loop ee =
+			match ee.eexpr with
+			| TBinop _ -> parent e2
+			| TCast (e,None) -> loop e
+			| _ -> e2
+		in
+		{ e with eexpr = TUnop (op,mode,loop e2) }
+	| TIf (e1,e2,eelse) ->
+		let e1 = (match e1.eexpr with
+			| TParenthesis _ -> e1
+			| _ -> parent e1
+		) in
+		let e2 = (match e2.eexpr, eelse with
+			| TIf (_,_,Some _) , _ | TIf (_,_,None), Some _ -> block e2
+			| _ -> e2
+		) in
+		{ e with eexpr = TIf (e1,e2,eelse) }
+	| TFunction f ->
+		(match f.tf_expr.eexpr with
+		| TBlock _ -> e
+		| _ -> { e with eexpr = TFunction { f with tf_expr = block f.tf_expr } })
+	| TCall (e2,args) ->		
+		if need_parent e2 then { e with eexpr = TCall(parent e2,args) } else e
+	| TField (e2,f) ->
+		if need_parent e2 then { e with eexpr = TField(parent e2,f) } else e
+	| TArray (e1,e2) ->
+		if need_parent e1 then { e with eexpr = TArray(parent e1,e2) } else e
+	| _ ->
+		e
+
+let reduce_expr ctx e =
+	match e.eexpr with
+	| TSwitch (_,cases,_) ->
+		List.iter (fun (cl,_) ->
+			List.iter (fun e ->
+				match e.eexpr with
+				| TCall ({ eexpr = TEnumField _ },_) -> error "Not-constant enum in switch cannot be matched" e.epos
+				| _ -> ()
+			) cl
+		) cases;
+		e
+	| TBlock [{ eexpr = TConst _ } as ec] ->
+		{ ec with epos = e.epos }
+	| TParenthesis ec ->
+		{ ec with epos = e.epos }
+	| _ ->
+		e
+
+let rec sanitize ctx e =
+	sanitize_expr (reduce_expr ctx (Type.map_expr (sanitize ctx) e))
+
 (* ---------------------------------------------------------------------- *)
 (* REDUCE *)
 
-let rec reduce_loop ctx is_sub e =
+let rec reduce_loop ctx e =
 	let is_float t =
 		match follow t with
 		| TInst ({ cl_path = ([],"Float") },_) -> true
 		| _ -> false
 	in
-	let e = Type.map_expr (reduce_loop ctx (match e.eexpr with TBlock _ -> false | _ -> true)) e in
-	match e.eexpr with
+	let e = Type.map_expr (reduce_loop ctx) e in
+	sanitize_expr (match e.eexpr with
 	| TIf ({ eexpr = TConst (TBool t) },e1,e2) ->
 		(if t then e1 else match e2 with None -> { e with eexpr = TBlock [] } | Some e -> e)
 	| TWhile ({ eexpr = TConst (TBool false) },sub,flag) ->
@@ -330,6 +482,11 @@ let rec reduce_loop ctx is_sub e =
 		| _ , TConst (TInt 1l) when op = OpMult -> e1
 		| _ , TConst (TFloat v) when (match op with OpAdd | OpSub -> float_of_string v = 0. && is_float e1.etype | _ -> false) -> e1 (* bits operations might cause overflow *)
 		| _ , TConst (TFloat v) when op = OpMult && float_of_string v = 1. && is_float e1.etype -> e1
+		| TConst TNull, TConst TNull ->
+			(match op with
+			| OpEq -> { e with eexpr = TConst (TBool true) }
+			| OpNotEq -> { e with eexpr = TConst (TBool false) }
+			| _ -> e)
 		| TConst (TInt a), TConst (TInt b) ->
 			let opt f = try { e with eexpr = TConst (TInt (f a b)) } with Exit -> e in
 			let check_overflow f =
@@ -413,7 +570,8 @@ let rec reduce_loop ctx is_sub e =
 			| OpAssign -> e
 			| _ ->
 				error "You cannot directly compare enums with arguments. Use either 'switch' or 'Type.enumEq'" e.epos)
-		| _ -> e)
+		| _ ->
+			e)
 	| TUnop (op,flag,esub) ->
 		(match op, esub.eexpr with
 		| Not, TConst (TBool f) -> { e with eexpr = TConst (TBool (not f)) }
@@ -431,54 +589,40 @@ let rec reduce_loop ctx is_sub e =
 	| TCall ({ eexpr = TFunction func } as ef,el) ->
 		(match follow ef.etype with
 		| TFun (_,rt) ->
-			let cf = { cf_name = ""; cf_params = []; cf_type = ef.etype; cf_public = true; cf_doc = None; cf_get = NormalAccess; cf_set = NoAccess; cf_expr = None } in
+			let cf = { cf_name = ""; cf_params = []; cf_type = ef.etype; cf_public = true; cf_doc = None; cf_meta = no_meta; cf_kind = Var { v_read = AccNormal; v_write = AccNo }; cf_expr = None } in
 			let inl = (try type_inline ctx cf func (mk (TConst TNull) (mk_mono()) e.epos) el rt e.epos with Error (Custom _,_) -> None) in
 			(match inl with
 			| None -> e
 			| Some e -> e)
-		| _ -> 
+		| _ ->
 			e)
-	| TParenthesis ({ eexpr = TConst _ } as ec) | TBlock [{ eexpr = TConst _ } as ec] ->
-		{ ec with epos = e.epos }
-	| TSwitch (_,cases,_) ->
-		List.iter (fun (cl,_) ->
-			List.iter (fun e ->
-				match e.eexpr with
-				| TCall ({ eexpr = TEnumField _ },_) -> error "Not-constant enum in switch cannot be matched" e.epos
-				| _ -> ()
-			) cl
-		) cases;
-		e
 	| _ ->
-		e
+		reduce_expr ctx e)
 
 let reduce_expression ctx e =
-	if ctx.com.foptimize then reduce_loop ctx false e else e
+	if ctx.com.foptimize then reduce_loop ctx e else e
 
 (* ---------------------------------------------------------------------- *)
-(* SANITIZE *)
+(* ELIMINATE DEAD CODE *)
 
 (*
-	makes sure that when an AST get generated to source code, it will not
-	generate expressions that evaluate differently. It is then necessary to
-	add parenthesises around some binary expressions when the AST does not
-	correspond to the natural operand priority order for the platform
+	if dead code elimination is on, any class without fields is eliminated from the output.
 *)
 
-let rec sanitize e =	
-	match e.eexpr with
-	| TBinop (op,e1,e2) ->
-		let parent e = mk (TParenthesis e) e.etype e.epos in
-		let e1 = sanitize e1 in
-		let e2 = sanitize e2 in
-		let e1 = (match e1.eexpr with
-			| TBinop (op2,_,_) when Parser.swap op2 op -> parent e1
-			| _ -> e1
-		) in
-		let e2 = (match e2.eexpr with
-			| TBinop (op2,_,_) when Parser.swap op2 op -> parent e2
-			| _ -> e2
-		) in
-		{ e with eexpr = TBinop (op,e1,e2) }
-	| _ -> Type.map_expr sanitize e
-
+let filter_dead_code com =
+	let s_class c = s_type_path c.cl_path in
+	com.types <- List.filter (fun t ->
+		match t with
+		| TClassDecl c ->
+			if (c.cl_extern or has_meta ":keep" c.cl_meta) then
+				true
+			else (
+				match (c.cl_ordered_statics, c.cl_ordered_fields, c.cl_constructor) with
+				| ([], [], None) ->
+					if com.verbose then print_endline ("Remove class " ^ s_class c);
+					false
+				| _ ->
+					true)
+		| _ ->
+			true
+	) com.types
