@@ -20,14 +20,29 @@ open Ast
 
 type path = string list * string
 
-type field_access =
-	| NormalAccess
-	| NoAccess
-	| ResolveAccess (* call resolve("field") when accessed *)
-	| CallAccess of string (* perform a method call when accessed *)
-	| MethodAccess of bool (* true = the method is dynamic *)
-	| InlineAccess (* similar to Normal but inline when acccessed *)
-	| NeverAccess (* can't be accessed, even in subclasses *)
+type field_kind =
+	| Var of var_kind
+	| Method of method_kind
+
+and var_kind = {
+	v_read : var_access;
+	v_write : var_access;
+}
+
+and var_access =
+	| AccNormal
+	| AccNo				(* can't be accessed outside of the class itself and its subclasses *)
+	| AccNever			(* can't be accessed, even in subclasses *)
+	| AccResolve		(* call resolve("field") when accessed *)
+	| AccCall of string (* perform a method call when accessed *)
+	| AccInline			(* similar to Normal but inline when accessed *)
+	| AccRequire of string (* set when @:require(cond) fails *)
+
+and method_kind =
+	| MethNormal
+	| MethInline
+	| MethDynamic
+	| MethMacro
 
 type t =
 	| TMono of t option ref
@@ -96,6 +111,7 @@ and texpr_expr =
 	| TBreak
 	| TContinue
 	| TThrow of texpr
+	| TCast of texpr * module_type option
 
 and texpr = {
 	eexpr : texpr_expr;
@@ -107,9 +123,9 @@ and tclass_field = {
 	cf_name : string;
 	mutable cf_type : t;
 	cf_public : bool;
-	cf_doc : Ast.documentation;
-	cf_get : field_access;
-	cf_set : field_access;
+	mutable cf_doc : Ast.documentation;
+	mutable cf_meta : metadata;
+	mutable cf_kind : field_kind;
 	cf_params : (string * t) list;
 	mutable cf_expr : texpr option;
 }
@@ -122,11 +138,14 @@ and tclass_kind =
 	| KGeneric
 	| KGenericInstance of tclass * tparams
 
+and metadata = Ast.metadata
+
 and tclass = {
-	cl_path : path;
-	cl_pos : Ast.pos;
-	cl_doc : Ast.documentation;
-	cl_private : bool;
+	mutable cl_path : path;
+	mutable cl_pos : Ast.pos;
+	mutable cl_private : bool;
+	mutable cl_doc : Ast.documentation;
+	mutable cl_meta : metadata;
 	mutable cl_kind : tclass_kind;
 	mutable cl_extern : bool;
 	mutable cl_interface : bool;
@@ -150,6 +169,7 @@ and tenum_field = {
 	ef_pos : Ast.pos;
 	ef_doc : Ast.documentation;
 	ef_index : int;
+	mutable ef_meta : metadata;
 }
 
 and tenum = {
@@ -157,7 +177,8 @@ and tenum = {
 	e_pos : Ast.pos;
 	e_doc : Ast.documentation;
 	e_private : bool;
-	e_extern : bool;
+	mutable e_meta : metadata;
+	mutable e_extern : bool;
 	mutable e_types : (string * t) list;
 	mutable e_constrs : (string , tenum_field) PMap.t;
 	mutable e_names : string list;
@@ -168,6 +189,7 @@ and tdef = {
 	t_pos : Ast.pos;
 	t_doc : Ast.documentation;
 	t_private : bool;
+	mutable t_meta : metadata;
 	mutable t_types : (string * t) list;
 	mutable t_type : t;
 }
@@ -199,12 +221,13 @@ let tfun pl r = TFun (List.map (fun t -> "",false,t) pl,r)
 
 let fun_args l = List.map (fun (a,c,t) -> a, c <> None, t) l
 
-let mk_class path pos doc priv =
+let mk_class path pos =
 	{
 		cl_path = path;
 		cl_pos = pos;
-		cl_doc = doc;
-		cl_private = priv;
+		cl_doc = None;
+		cl_meta = [];
+		cl_private = false;
 		cl_kind = KNormal;
 		cl_extern = false;
 		cl_interface = false;
@@ -222,7 +245,10 @@ let mk_class path pos doc priv =
 		cl_overrides = [];
 	}
 
-let null_class = mk_class ([],"") Ast.null_pos None true
+let null_class =
+	let c = mk_class ([],"") Ast.null_pos in
+	c.cl_private <- true;
+	c
 
 let arg_name (name,_,_) = name
 
@@ -236,6 +262,11 @@ let t_path = function
 	| TEnumDecl e -> e.e_path
 	| TTypeDecl t -> t.t_path
 
+let t_pos = function
+	| TClassDecl c -> c.cl_pos
+	| TEnumDecl e -> e.e_pos
+	| TTypeDecl t -> t.t_pos
+
 let print_context() = ref []
 
 let is_closed a = !(a.a_status) <> Opened
@@ -286,13 +317,23 @@ and s_type_params ctx = function
 	| l -> "<" ^ String.concat ", " (List.map (s_type ctx) l) ^ ">"
 
 let s_access = function
-	| NormalAccess -> "default"
-	| NoAccess -> "null"
-	| NeverAccess -> "never"
-	| CallAccess m -> m
-	| MethodAccess b -> if b then "dynamic method" else "default method"
-	| ResolveAccess -> "resolve"
-	| InlineAccess -> "inline"
+	| AccNormal -> "default"
+	| AccNo -> "null"
+	| AccNever -> "never"
+	| AccResolve -> "resolve"
+	| AccCall m -> m
+	| AccInline	-> "inline"
+	| AccRequire n -> "require " ^ n
+
+let s_kind = function
+	| Var { v_read = AccNormal; v_write = AccNormal } -> "var"
+	| Var v -> "(" ^ s_access v.v_read ^ "," ^ s_access v.v_write ^ ")"
+	| Method m ->
+		match m with
+		| MethNormal -> "method"
+		| MethDynamic -> "dynamic method"
+		| MethInline -> "inline method"
+		| MethMacro -> "macro method"
 
 let rec is_parent csup c =
 	if c == csup then
@@ -301,45 +342,6 @@ let rec is_parent csup c =
 		| None -> false
 		| Some (c,_) -> is_parent csup c
 
-let rec link e a b =
-	(* tell if a is is b *)
-	let rec loop t =
-		if t == a then
-			true
-		else match t with
-		| TMono t -> (match !t with None -> false | Some t -> loop t)
-		| TEnum (e,tl) -> e.e_path = ([],"Protected") || List.exists loop tl
-		| TInst (_,tl) | TType (_,tl) -> List.exists loop tl
-		| TFun (tl,t) -> List.exists (fun (_,_,t) -> loop t) tl || loop t
-		| TDynamic t2 ->
-			if t == t2 then
-				false
-			else
-				loop t2
-		| TLazy f ->
-			loop (!f())
-		| TAnon a ->
-			try
-				PMap.iter (fun _ f -> if loop f.cf_type then raise Exit) a.a_fields;
-				false
-			with
-				Exit -> true
-	in
-	(* tell if a ~= b *)
-	let rec loop2 t =
-		if t == a then
-			true
-		else match t with
-		| TMono t -> (match !t with None -> false | Some t -> loop2 t)
-		| _ -> false
-	in
-	if loop b then
-		loop2 b
-	else
-		match b with
-		| TDynamic _ -> true
-		| _ -> e := Some b; true
-
 let map loop t =
 	match t with
 	| TMono r ->
@@ -447,6 +449,38 @@ let rec follow t =
 		follow (apply_params t.t_types tl t.t_type)
 	| _ -> t
 
+let rec link e a b =
+	(* tell if setting a == b will create a type-loop *)
+	let rec loop t =
+		if t == a then
+			true
+		else match t with
+		| TMono t -> (match !t with None -> false | Some t -> loop t)
+		| TEnum (_,tl) -> List.exists loop tl
+		| TInst (_,tl) | TType (_,tl) -> List.exists loop tl
+		| TFun (tl,t) -> List.exists (fun (_,_,t) -> loop t) tl || loop t
+		| TDynamic t2 ->
+			if t == t2 then
+				false
+			else
+				loop t2
+		| TLazy f ->
+			loop (!f())
+		| TAnon a ->
+			try
+				PMap.iter (fun _ f -> if loop f.cf_type then raise Exit) a.a_fields;
+				false
+			with
+				Exit -> true
+	in
+	(* tell is already a ~= b *)
+	if loop b then
+		(follow b) == a
+	else
+		match b with
+		| TDynamic _ -> true
+		| _ -> e := Some b; true
+
 let monomorphs eparams t =
 	apply_params eparams (List.map (fun _ -> mk_mono()) eparams) t
 
@@ -475,7 +509,7 @@ type unify_error =
 	| Invalid_field_type of string
 	| Has_no_field of t * string
 	| Has_extra_field of t * string
-	| Invalid_access of string * bool * field_access * field_access
+	| Invalid_kind of string * field_kind * field_kind
 	| Invalid_visibility of string
 	| Not_matching_optional of string
 	| Cant_force_optional
@@ -484,35 +518,54 @@ exception Unify_error of unify_error list
 
 let cannot_unify a b = Cannot_unify (a,b)
 let invalid_field n = Invalid_field_type n
-let invalid_access n get a b = Invalid_access (n,get,a,b)
+let invalid_kind n a b = Invalid_kind (n,a,b)
 let invalid_visibility n = Invalid_visibility n
 let has_no_field t n = Has_no_field (t,n)
 let has_extra_field t n = Has_extra_field (t,n)
 let error l = raise (Unify_error l)
-
-type simple_access =
-	| SAYes
-	| SANo
-	| SARuntime
-
-let simple_access = function
-	| NormalAccess | InlineAccess | MethodAccess true -> SAYes
-	| NoAccess | NeverAccess | MethodAccess false -> SANo
-	| ResolveAccess | CallAccess _ -> SARuntime
+let has_meta m ml = List.exists (fun (m2,_,_) -> m = m2) ml
+let no_meta = []
 
 (*
 	we can restrict access as soon as both are runtime-compatible
 *)
 let unify_access a1 a2 =
-	a1 = a2 || match simple_access a1 , simple_access a2 with
-		| SAYes, SAYes
-		| _, SANo -> true
-		| _ -> false
+	a1 = a2 || match a1, a2 with
+	| _, AccNo | _, AccNever -> true
+	| AccInline, AccNormal -> true
+	| _ -> false
+
+let direct_access = function
+	| AccNo | AccNever | AccNormal | AccInline | AccRequire _ -> true
+	| AccResolve | AccCall _ -> false
+
+let unify_kind k1 k2 =
+	k1 = k2 || match k1, k2 with
+		| Var v1, Var v2 -> unify_access v1.v_read v2.v_read && unify_access v1.v_write v2.v_write
+		| Var v, Method m ->
+			(match v.v_read, v.v_write, m with
+			| AccNormal, _, MethNormal -> true
+			| AccNormal, AccNormal, MethDynamic -> true
+			| _ -> false)
+		| Method m, Var v ->
+			(match m with
+			| MethDynamic -> direct_access v.v_read && direct_access v.v_write
+			| MethMacro -> false
+			| MethNormal | MethInline ->
+				match v.v_write with
+				| AccNo | AccNever -> true
+				| _ -> false)
+		| Method m1, Method m2 ->
+			match m1,m2 with
+			| MethInline, MethNormal
+			| MethDynamic, MethNormal -> true
+			| _ -> false
 
 let eq_stack = ref []
 
 type eq_kind =
 	| EqStrict
+	| EqCoreType
 	| EqRightDynamic
 	| EqBothDynamic
 
@@ -524,16 +577,17 @@ let rec type_eq param a b =
 	| _ , TLazy f -> type_eq param a (!f())
 	| TMono t , _ ->
 		(match !t with
-		| None -> if not (link t a b) then error [cannot_unify a b]
+		| None -> if param = EqCoreType || not (link t a b) then error [cannot_unify a b]
 		| Some t -> type_eq param t b)
 	| _ , TMono t ->
 		(match !t with
-		| None -> if not (link t b a) then error [cannot_unify a b]
+		| None -> if param = EqCoreType || not (link t b a) then error [cannot_unify a b]
 		| Some t -> type_eq param a t)
-	| TType (t1,tl1), TType (t2,tl2) when t1 == t2 && List.length tl1 = List.length tl2 ->
+	| TType (t1,tl1), TType (t2,tl2) when (t1 == t2 || (param = EqCoreType && t1.t_path = t2.t_path)) && List.length tl1 = List.length tl2 ->
 		List.iter2 (type_eq param) tl1 tl2
-	| TType (t,tl) , _ -> type_eq param (apply_params t.t_types tl t.t_type) b
-	| _ , TType (t,tl) ->
+	| TType (t,tl) , _ when param <> EqCoreType ->
+		type_eq param (apply_params t.t_types tl t.t_type) b
+	| _ , TType (t,tl) when param <> EqCoreType ->
 		if List.exists (fun (a2,b2) -> fast_eq a a2 && fast_eq b b2) (!eq_stack) then
 			()
 		else begin
@@ -547,10 +601,10 @@ let rec type_eq param a b =
 					error (cannot_unify a b :: l)
 		end
 	| TEnum (e1,tl1) , TEnum (e2,tl2) ->
-		if e1 != e2 then error [cannot_unify a b];
+		if e1 != e2 && not (param = EqCoreType && e1.e_path = e2.e_path) then error [cannot_unify a b];
 		List.iter2 (type_eq param) tl1 tl2
 	| TInst (c1,tl1) , TInst (c2,tl2) ->
-		if c1 != c2 then error [cannot_unify a b];
+		if c1 != c2 && not (param = EqCoreType && c1.cl_path = c2.cl_path) then error [cannot_unify a b];
 		List.iter2 (type_eq param) tl1 tl2
 	| TFun (l1,r1) , TFun (l2,r2) when List.length l1 = List.length l2 ->
 		(try
@@ -568,8 +622,7 @@ let rec type_eq param a b =
 			PMap.iter (fun n f1 ->
 				try
 					let f2 = PMap.find n a2.a_fields in
-					if f1.cf_get <> f2.cf_get && (param = EqStrict || not (unify_access f1.cf_get f2.cf_get)) then error [invalid_access n true f1.cf_get f2.cf_get];
-					if f1.cf_set <> f2.cf_set && (param = EqStrict || not (unify_access f1.cf_set f2.cf_set)) then error [invalid_access n false f1.cf_set f2.cf_set];
+					if f1.cf_kind <> f2.cf_kind && (param = EqStrict || param = EqCoreType || not (unify_kind f1.cf_kind f2.cf_kind)) then error [invalid_kind n f1.cf_kind f2.cf_kind];
 					try
 						type_eq param f1.cf_type f2.cf_type
 					with
@@ -652,10 +705,16 @@ let rec unify a b =
 		| None -> if not (link t b a) then error [cannot_unify a b]
 		| Some t -> unify a t)
 	| TType (t,tl) , _ ->
-		(try
-			unify (apply_params t.t_types tl t.t_type) b
-		with
-			Unify_error l -> error (cannot_unify a b :: l))
+		if not (List.exists (fun (a2,b2) -> fast_eq a a2 && fast_eq b b2) (!unify_stack)) then begin
+			try
+				unify_stack := (a,b) :: !unify_stack;
+				unify (apply_params t.t_types tl t.t_type) b;
+				unify_stack := List.tl !unify_stack;
+			with
+				Unify_error l ->
+					unify_stack := List.tl !unify_stack;
+					error (cannot_unify a b :: l)
+		end
 	| _ , TType (t,tl) ->
 		if not (List.exists (fun (a2,b2) -> fast_eq a a2 && fast_eq b b2) (!unify_stack)) then begin
 			try
@@ -697,8 +756,7 @@ let rec unify a b =
 		(try
 			PMap.iter (fun n f2 ->
 				let ft, f1 = (try class_field c n with Not_found -> error [has_no_field a n]) in
-				if not (unify_access f1.cf_get f2.cf_get) then error [invalid_access n true f1.cf_get f2.cf_get];
-				if not (unify_access f1.cf_set f2.cf_set) then error [invalid_access n false f1.cf_set f2.cf_set];
+				if not (unify_kind f1.cf_kind f2.cf_kind) then error [invalid_kind n f1.cf_kind f2.cf_kind];
 				if f2.cf_public && not f1.cf_public then error [invalid_visibility n];
 				try
 					unify_with_access (apply_params c.cl_types tl ft) f2
@@ -713,8 +771,7 @@ let rec unify a b =
 			PMap.iter (fun n f2 ->
 			try
 				let f1 = PMap.find n a1.a_fields in
-				if not (unify_access f1.cf_get f2.cf_get) then error [invalid_access n true f1.cf_get f2.cf_get];
-				if not (unify_access f1.cf_set f2.cf_set) then error [invalid_access n false f1.cf_set f2.cf_set];
+				if not (unify_kind f1.cf_kind f2.cf_kind) then error [invalid_kind n f1.cf_kind f2.cf_kind];
 				if f2.cf_public && not f1.cf_public then error [invalid_visibility n];
 				try
 					unify_with_access f1.cf_type f2;
@@ -793,11 +850,14 @@ and unify_types a b tl1 tl2 =
 	with
 		Unify_error l -> error ((cannot_unify a b) :: l)
 
-and unify_with_access t f =
-	match f.cf_get, f.cf_set with
-	| NoAccess , _ | NeverAccess, _ -> unify f.cf_type t
-	| _ , NoAccess | _, NeverAccess -> unify t f.cf_type
-	| _ , _ -> type_eq EqBothDynamic t f.cf_type
+and unify_with_access t1 f2 =
+	match f2.cf_kind with
+	(* write only *)
+	| Var { v_read = AccNo } | Var { v_read = AccNever } -> unify f2.cf_type t1
+	(* read only *)
+	| Method MethNormal | Method MethInline | Var { v_write = AccNo } | Var { v_write = AccNever } -> unify t1 f2.cf_type
+	(* read/write *)
+	| _ -> type_eq EqBothDynamic t1 f2.cf_type
 
 let iter f e =
 	match e.eexpr with
@@ -818,6 +878,7 @@ let iter f e =
 	| TField (e,_)
 	| TClosure (e,_)
 	| TParenthesis e
+	| TCast (e,_)
 	| TUnop (_,_,e) ->
 		f e
 	| TArrayDecl el
@@ -902,6 +963,8 @@ let map_expr f e =
 		{ e with eexpr = TTry (f e1, List.map (fun (v,t,e) -> v, t, f e) catches) }
 	| TReturn eo ->
 		{ e with eexpr = TReturn (match eo with None -> None | Some e -> Some (f e)) }
+	| TCast (e1,t) ->
+		{ e with eexpr = TCast (f e1,t) }
 
 let map_expr_type f ft e =
 	match e.eexpr with
@@ -962,6 +1025,39 @@ let map_expr_type f ft e =
 		{ e with eexpr = TTry (f e1, List.map (fun (v,t,e) -> v, ft t, f e) catches); etype = ft e.etype }
 	| TReturn eo ->
 		{ e with eexpr = TReturn (match eo with None -> None | Some e -> Some (f e)); etype = ft e.etype }
+	| TCast (e1,t) ->
+		{ e with eexpr = TCast (f e1,t); etype = ft e.etype }
+
+let s_expr_kind e =
+	match e.eexpr with
+	| TConst _ -> "Const"
+	| TLocal _ -> "Local"
+	| TEnumField _ -> "EnumField"
+	| TArray (_,_) -> "Array"
+	| TBinop (_,_,_) -> "Binop"
+	| TField (_,_) -> "Field"
+	| TClosure _ -> "Closure"
+	| TTypeExpr _ -> "TypeExpr"
+	| TParenthesis _ -> "Parenthesis"
+	| TObjectDecl _ -> "ObjectDecl"
+	| TArrayDecl _ -> "ArrayDecl"
+	| TCall (_,_) -> "Call"
+	| TNew (_,_,_) -> "New"
+	| TUnop (_,_,_) -> "Unop"
+	| TFunction _ -> "Function"
+	| TVars _ -> "Vars"
+	| TBlock _ -> "Block"
+	| TFor (_,_,_,_) -> "For"
+	| TIf (_,_,_) -> "If"
+	| TWhile (_,_,_) -> "While"
+	| TSwitch (_,_,_) -> "Switch"
+	| TMatch (_,_,_,_) -> "Match"
+	| TTry (_,_) -> "Try"
+	| TReturn _ -> "Return"
+	| TBreak -> "Break"
+	| TContinue -> "Continue"
+	| TThrow _ -> "Throw"
+	| TCast _ -> "Cast"
 
 let rec s_expr s_type e =
 	let sprintf = Printf.sprintf in
@@ -1040,5 +1136,7 @@ let rec s_expr s_type e =
 		"Continue"
 	| TThrow e ->
 		"Throw " ^ (loop e)
+	| TCast (e,t) ->
+		sprintf "Cast %s%s" (match t with None -> "" | Some t -> s_type_path (t_path t) ^ ": ") (loop e)
 	) in
 	sprintf "(%s : %s)" str (s_type e.etype)
