@@ -78,6 +78,19 @@ type as3_op =
 	| A3OIAdd
 	| A3OISub
 	| A3OIMul
+	| A3OMemSet8
+	| A3OMemSet16
+	| A3OMemSet32
+	| A3OMemSetFloat
+	| A3OMemSetDouble
+	| A3OMemGet8
+	| A3OMemGet16
+	| A3OMemGet32
+	| A3OMemGetFloat
+	| A3OMemGetDouble
+	| A3OSign1
+	| A3OSign8
+	| A3OSign16
 
 type as3_name = as3_multi_name index
 
@@ -87,6 +100,8 @@ and as3_opcode =
 	| A3Throw
 	| A3GetSuper of as3_name
 	| A3SetSuper of as3_name
+	| A3DxNs of as3_ident index
+	| A3DxNsLate
 	| A3RegKill of reg
 	| A3Label
 	| A3Jump of as3_jump * int
@@ -127,10 +142,12 @@ and as3_opcode =
 	| A3CallPropLex of as3_name * nargs
 	| A3CallSuperVoid of as3_name * nargs
 	| A3CallPropVoid of as3_name * nargs
+	| A3ApplyType of nargs
 	| A3Object of nargs
 	| A3Array of nargs
 	| A3NewBlock
 	| A3ClassDef of unit index_nz
+	| A3GetDescendants of as3_name
 	| A3Catch of int
 	| A3FindPropStrict of as3_name
 	| A3FindProp of as3_name
@@ -195,6 +212,8 @@ and as3_multi_name =
 	| A3MRuntimeNameLate
 	| A3MMultiNameLate of as3_ns_set index
 	| A3MAttrib of as3_multi_name
+	| A3MParams of as3_multi_name index * as3_multi_name index list
+	| A3MAny
 
 and as3_value =
 	| A3VNone
