@@ -131,6 +131,8 @@ let rec as3_name_length t =
 		idx_opt_length id + idx_length r
 	| A3MName (id,r) ->
 		idx_length r + idx_length id
+	| A3MAny ->
+		int_length 0 + int_length 0
 	| A3MRuntimeName i ->
 		idx_length i
 	| A3MRuntimeNameLate ->
@@ -327,8 +329,13 @@ let rec read_name ctx ?k ch =
 	let k = (match k with None -> IO.read_byte ch | Some k -> k) in
 	match k with
 	| 0x07 ->
-		let ns = index ctx.as3_namespaces (read_int ch) in
-		let id = index ctx.as3_idents (read_int ch) in
+		let i = read_int ch in
+		let j = read_int ch in
+		if i = 0 && j = 0 then
+			A3MAny
+		else
+		let ns = index ctx.as3_namespaces i in
+		let id = index ctx.as3_idents j in
 		(* both ns and id can be 0 <=> '*' *)
 		A3MName (id,ns)
 	| 0x09 ->
@@ -561,8 +568,7 @@ let read_function ctx ch =
 let header_magic = 0x002E0010
 
 let parse ch len =
-	let data = IO.nread ch len in
-	let ch = IO.input_string data in
+	let ch, get_pos = IO.pos_in ch in
 	if IO.read_i32 ch <> header_magic then assert false;
 	let ints = read_list ch read_as3_int in
 	let uints = read_list ch read_as3_uint in
@@ -593,11 +599,11 @@ let parse ch len =
 	if parse_statics then ctx.as3_statics <- Array.map (fun _ -> read_static ctx ch) ctx.as3_classes;
 	if parse_inits then ctx.as3_inits <- read_list2 ch (read_static ctx);
 	if parse_functions then ctx.as3_functions <- read_list2 ch (read_function ctx);
-	ctx.as3_unknown <- IO.read_all ch;
+	ctx.as3_unknown <- IO.really_nread ch (len - (get_pos()));
 	if parse_functions && String.length ctx.as3_unknown <> 0 then assert false;
-	let len2 = as3_length ctx in
+(*	let len2 = as3_length ctx in
 	if len2 <> len then begin Printf.printf "%d != %d" len len2; assert false; end;
-	ctx
+*)	ctx
 
 (* ************************************************************************ *)
 (* WRITING *)
@@ -691,6 +697,10 @@ let rec write_name ch ?k x =
 		IO.write_byte ch (b 0x07);
 		write_index ch r;
 		write_index ch id
+	| A3MAny ->
+		IO.write_byte ch (b 0x07);
+		write_int ch 0;
+		write_int ch 0;
 	| A3MRuntimeName i ->
 		IO.write_byte ch (b 0x0F);
 		write_index ch i
@@ -701,7 +711,7 @@ let rec write_name ch ?k x =
 		write_index ch id
 	| A3MAttrib n ->
 		write_name ch ~k:(match n with
-			| A3MName _ -> 0x0D
+			| A3MName _ | A3MAny -> 0x0D
 			| A3MMultiName _ -> 0x0E
 			| A3MRuntimeName _ -> 0x10
 			| A3MRuntimeNameLate -> 0x12
@@ -854,7 +864,7 @@ let write_function ch f =
 	write_list2 ch write_field f.fun3_locals
 
 let write ch1 ctx =
-	let ch = IO.output_string() in
+	let ch = IO.output_strings() in
 	let empty_index = as3_empty_index ctx in
 	IO.write_i32 ch header_magic;
 	write_list ch write_as3_int ctx.as3_ints;
@@ -872,7 +882,7 @@ let write ch1 ctx =
 	if parse_functions then write_list2 ch write_function ctx.as3_functions;
 	IO.nwrite ch ctx.as3_unknown;
 	let str = IO.close_out ch in
-	IO.nwrite ch1 str
+	List.iter (IO.nwrite ch1) str
 
 (* ************************************************************************ *)
 (* DUMP *)
@@ -903,6 +913,7 @@ let ns_set_str ctx i =
 let rec name_str ctx kind t =
 	let rec loop = function
 		| A3MName (id,r) -> Printf.sprintf "%s %s%s" (namespace_str ctx r) kind (ident_str ctx id)
+		| A3MAny -> "ANY"
 		| A3MMultiName (id,r) -> Printf.sprintf "[%s %s%s]" (ns_set_str ctx r) kind (match id with None -> "NO" | Some i -> ident_str ctx i)
 		| A3MRuntimeName id -> Printf.sprintf "'%s'" (ident_str ctx id)
 		| A3MRuntimeNameLate -> "RTLATE"
@@ -936,7 +947,7 @@ let method_str ?(infos=false) ctx m =
 	(String.concat ", " (List.map (fun a ->
 		let id = (match m.mt3_pnames with
 			| None -> "p" ^ string_of_int !pcount
-			| Some l -> 
+			| Some l ->
 				match List.nth l !pcount with
 				| None -> "p" ^ string_of_int !pcount
 				| Some i -> ident_str ctx i
